/* $Id: stepper.c,v 0.04 2004/6/11 17:13 root Exp root $
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 *  stepper.c: Steppermotor driver with L6219 (ST-microelectronics)
 *
 *  Written by ps@telin.ugent.be (Philippe Serbruyns)
 *
 */
#include <asm/io.h>
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <sched.h>
#include <time.h>
#include <stdlib.h>
#include "stepper.h"

extern int ioperm (unsigned long from, unsigned long num, int turn_on);

unsigned char clockwise; /* turning direction */
double N1; /* initial rotationspeed (RPM, rounds per minute) */
const double tau=1.0; /* time constant in seconds, larger is slower */
const double transient_time=5.0*1.0; /* transient time */
const double steps_per_turn=200.0; /* motor specific */
char verbose=0;

/* states corrected by Massimo Manca to reflect the scheme */
const unsigned char state[]={0000,0004,0044,0040};
const unsigned char powerset[]={0033,0022,0011,0000};
unsigned char power;
long next_step=0;

/* use nanosleep(x) for delays < 2 ms
 * use usleep(x) for delay > 20 ms, also add 10 ms overhead
 * use sleep(x) for delay > 1 sec
 */

void delay(double delay)
{ static struct timespec wait;
 wait.tv_sec=0;
/* use nanosleep to extend delay precision */
if (delay>1e-3)
     do { wait.tv_nsec=1e6; nanosleep(&wait,NULL);
	  delay-=1e-3;
         } while (delay>1e-3);
wait.tv_nsec=delay*1e9;
nanosleep(&wait,NULL);
}

void step(double ndelay)
{
 if (clockwise) next_step++; else next_step--;
  outb(state[next_step&3]|powerset[power],LP_B); delay(ndelay);
 if(verbose>2) printf("step=%ld d=%f\n",next_step,ndelay);
}

void rotate(double N2)
{ static double N,t,d;
  static int i,trans;
  static char up;
/* N1 start RPM 
 * N2 final RPM
 */
up=1; N=N2-N1; 
if (N<0) { N=-N; up=0; }
if (verbose) printf("%s motor %s from %4.1f to %4.1f RPM (power=%d)\n",
	up?"Speeding":"Slowing",
	clockwise?"clockwise":"counter clockwise",N1,N2,power);
trans=transient_time*N;i=trans-1;
do {
  if (up) { t=(double)(trans-i)/N;
	    d=60.0/(steps_per_turn*(N*(1.0-exp(-t/tau))+N1)); }
  else { t=(double)i/N;
         d=60.0/(steps_per_turn*(N*(1.0-exp(-t/tau))+N2)); }
  step(d); 
  } while (--i>0);
N1=N2; /* succeeded, final RPM */
if(verbose>1) printf("steps=%ld\n",next_step);
}

void turns(float turns)
{ static double d;
  static long i;
d=60.0/steps_per_turn/N1;
i=turns*steps_per_turn;
if (verbose) 
   printf("Engine %s at %4.1f RPM for %4.1f turns (%ld steps,power=%d)\n",
           clockwise?"clockwise":"counter clockwise",N1,turns,i,power);
do { step(d); } while (--i>0);
if(verbose>1) printf("steps=%ld\n",next_step);
}

void jump(float newspeed)
{ N1=newspeed;
 outb(state[next_step&3]|powerset[power],LP_B);
 if (verbose) printf("Stepper ready for jumping %s to %4.1f RPM (power=%d)\n",
              clockwise?"clockwise":"counter clockwise",N1,power);
}

void wait(double time)
{  outb(state[next_step&3]|powerset[power],LP_B);
if (verbose) printf("Waiting %2.1f seconds (power=%d)\n",time,power);
 if (time>0.999999) sleep(floor(time));
 if (time-floor(time)>20e-3) delay(time-floor(time));
}
 

void main (int argc, char *argv[])
{ int pid;
  struct sched_param p;
  char token[80];

ioperm (LP_B, 3L, 1);      /* release io-addresses LP_B,LP_S and LP_C */

/* change round ribbon scheduler to request minimal interrupts */
pid=getpid();
p.sched_priority=10; /* set priority 0 < x < 100 */
sched_setscheduler(pid,SCHED_FIFO, &p);

N1=0.0; /* stepper stands still at start */
/* interprete the PSDL (Portable Stepper Discription Language) */
while (scanf("%s",token)!=EOF) switch(token[0]) {
/* skip comments */
case '#': while (scanf("%c",token)!=EOF)
           if (token[0]=='\n') break;
          break;

/* ClockWise or Counter ClockWise */
case 'C': if (token[1]=='W') clockwise=1; else clockwise=0;
     break;

/* Set power 0|1|2|3 */
case 'P': power=atoi(token+1);
     break;

/* change rotation speed, units in RPM */
case 'R': rotate(atof(token+1));
     break;

/* jump to desired rotation speed, units in RPM */
case 'J': jump(atof(token+1));
     break;

/* motor turns N turns (N=float number) */
case 'T': turns(atof(token+1));
     break;

/* wait N seconds  (N=float number) */
case 'W': wait(atof(token+1));
     break;

/* set verbose level
 * V0 -> nothing printed, optimized for speedy operations
 * V1 -> inform speeds
 * V2 -> for each step shows delay
 */
case 'V': verbose=atoi(token+1);
     break;
 }
}
